#!/usr/bin/env python3

'''
Linode external inventory script
================================

Generates inventory that Ansible can understand by making API request to
Linode API4.

You need /etc/linode.cfg with the token of your account

For more details, see: https://linode-api4.readthedocs.io

NOTE: This script also assumes that the Linodes in your account all have
labels that correspond to hostnames that are in your resolver search path.
Your resolver search path resides in /etc/hosts.
'''

try:
    import configparser
except:
    import ConfigParser as configparser

import argparse
# Standard imports
import os
import re
import sys
from ansible.errors import AnsibleError as ae
from time import time

try:
    import json
except ImportError:
    import simplejson as json

from linode_api4 import LinodeClient
from linode_api4.errors import ApiError

class LinodeInventory(object):

    def _empty_inventory(self):
        return {"_meta": {"hostvars": {}}}

    def __init__(self):
        """Main execution path."""
        # Inventory grouped by display group
        self.inventory = self._empty_inventory()

        # Index of label to Linode ID
        self.index = {}

        # Local cache of Datacenter objects populated by populate_datacenter_cache()
        self._datacenter_cache = None

        # Read settings and parse CLI arguments
        self.parse_cli_args()
        self.read_settings()
        self.read_config()

        # Cache
        if self.args.refresh_cache:
            self.do_api_calls_update_cache()
        elif not self.is_cache_valid():
            self.do_api_calls_update_cache()

        # Data to print
        if self.args.host:
            data_to_print = self.get_host_info()
        elif self.args.list:
            # Display list of nodes for inventory
            if self.inventory == self._empty_inventory():
                data_to_print = self.get_inventory_from_cache()
            else:
                data_to_print = self.json_format_dict(self.inventory, True)
        else:
            data_to_print = {}

        print(data_to_print)

    def parse_cli_args(self):
        """Command line argument processing"""
        parser = argparse.ArgumentParser(description='Produce an Ansible Inventory file based on Linode')
        parser.add_argument('--company', default=os.environ.get('ANSIBLE_COMPANY', None),
                            help='Only useful if inventories from multiple companies are installed')
        parser.add_argument('--list', action='store_true', default=True,
                           help='List nodes (default: True)')
        parser.add_argument('--host', action='store',
                           help='Get all the variables about a specific node')
        parser.add_argument('--refresh-cache', action='store_true', default=os.environ.get('ANSIBLE_INVENTORY_REFRESH', False),
                           help='Force refresh of cache by making API requests to Linode (default: False - use cache files)')
        self.args = parser.parse_args()

    def read_settings(self):
        ''' Reads the settings from the linode.ini file '''
        config = configparser.ConfigParser()
        linode_default_ini_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'linode.ini')
        linode_ini_path = os.path.expanduser(os.path.expandvars(os.environ.get('LINODE_INI_PATH', linode_default_ini_path)))
        config.read(linode_ini_path)

        # Cache related
        cache_dir = os.path.expanduser(config.get('linode', 'cache_path'))
        if not os.path.exists(cache_dir):
            os.makedirs(cache_dir)
        suffix = ''
        if self.args.company:
            suffix = '-' + self.args.company
        self.cache_path_cache = cache_dir + "/ansible-linode" + suffix + ".cache"
        self.cache_path_index = cache_dir + "/ansible-linode" + suffix + ".index"
        self.cache_max_age = config.getint('linode', 'cache_max_age')

    def read_config(self):
        configPath = '/etc/linode.cfg'
        config = configparser.ConfigParser()
        config.read(configPath)

        if self.args.company:
            section = 'profile ' + self.args.company
        else:
            section = 'Credentials'
        if config.has_option(section, 'token'):
            self.client = LinodeClient(config.get(section, 'token'))
        else:
            raise ae('Can not find credentials')

    def is_cache_valid(self):
        ''' Determines if the cache files have expired, or if it is still valid '''

        if os.path.isfile(self.cache_path_cache):
            mod_time = os.path.getmtime(self.cache_path_cache)
            current_time = time()
            if (mod_time + self.cache_max_age) > current_time:
                if os.path.isfile(self.cache_path_index):
                    return True

        return False

    def do_api_calls_update_cache(self):
        """Do API calls, and save data in cache files."""
        self.get_nodes()
        self.write_to_cache(self.inventory, self.cache_path_cache)
        self.write_to_cache(self.index, self.cache_path_index)

    def get_nodes(self):
        """Makes an Linode API call to get the list of nodes."""
        try:
            for node in self.client.linode.instances():
                self.add_node(node)
        except ApiError as e:
            print("Looks like Linode's API is down:")
            print("")
            print(e)
            sys.exit(1)

    def get_node(self, linode_id):
        """Gets details about a specific node."""
        try:
            return self.client.load(self.client.Instance, linode_id)
        except ApiError as e:
            print("Looks like Linode's API is down:")
            print("")
            print(e)
            sys.exit(1)

    def populate_datacenter_cache(self):
        """Creates self._datacenter_cache, containing all Datacenters indexed by ID."""
        self._datacenter_cache = {}
        #dcs = Datacenter.search()
        #for dc in dcs:
        #    self._datacenter_cache[dc.api_id] = dc

    def get_datacenter_city(self, node):
        """Returns a the lowercase city name of the node's data center."""
        if self._datacenter_cache is None:
            self.populate_datacenter_cache()
        location = self._datacenter_cache[node.datacenter_id].location
        location = location.lower()
        location = location.split(",")[0]
        return location

    def add_node(self, node):
        """Adds an node to the inventory and index."""

        nodeMeta = self.get_host_info_dict_from_instance(node);
        if nodeMeta['linode_status'] == 'offline':
            return

        dest = node.label

        # Add to index
        id = 'l%s' % node.id
        self.index[dest] = id

        # Inventory: Group by node ID (always a group of 1)
        self.inventory[id] = [dest]

        # Inventory: Group by datacenter city
        # self.push(self.inventory, self.get_datacenter_city(node), dest)

        self.expand_csv_tags = True
        # Inventory: Group by dipslay group
        v = node.group
        if self.expand_csv_tags and v and ',' in v:
            values = []
            for x in v.split(','):
                values.append(x.strip())
        else:
            values = [v]
        for v in values:
            self.push(self.inventory, self.to_safe(v), dest)

        # Global Tag: tag all JiffyBox instances
        self.push(self.inventory, 'linode', dest)

        self.inventory["_meta"]["hostvars"][dest] = nodeMeta

    def get_host_info_dict_from_instance(self, node):
        instance_vars = {}
        for direct_attr in [
            "id",
            "label",
            "status",
            "group"
        ]:
            instance_vars["linode_" + direct_attr] = getattr(node, direct_attr)

        # instance_vars["linode_datacenter_city"] = self.get_datacenter_city(node)

        instance_vars["linode_public_ips"] = [addr.address for addr in node.ips.ipv4.public]
        instance_vars["linode_private_ips"] = [addr.address for addr in node.ips.ipv4.private]

        instance_vars['static_ipv4'] = instance_vars['linode_public_ips'][0]
        return instance_vars

    def get_host_info(self):
        """Get variables about a specific host."""

        if len(self.index) == 0:
            # Need to load index from cache
            self.load_index_from_cache()

        if not self.args.host in self.index:
            # try updating the cache
            self.do_api_calls_update_cache()
            if not self.args.host in self.index:
                # host might not exist anymore
                return self.json_format_dict({}, True)

        node_id = self.index[self.args.host]

        node = self.get_node(node_id)
        return self.json_format_dict(self.get_host_info_dict_from_instance(node), True)

    def push(self, my_dict, key, element):
        """Pushed an element onto an array that may not have been defined in the dict."""
        if key in my_dict:
            my_dict[key].append(element)
        else:
            my_dict[key] = [element]

    def get_inventory_from_cache(self):
        """Reads the inventory from the cache file and returns it as a JSON object."""
        cache = open(self.cache_path_cache, 'r')
        json_inventory = cache.read()
        return json_inventory

    def load_index_from_cache(self):
        """Reads the index from the cache file and sets self.index."""
        cache = open(self.cache_path_index, 'r')
        json_index = cache.read()
        self.index = json.loads(json_index)

    def write_to_cache(self, data, filename):
        """Writes data in JSON format to a file."""
        json_data = self.json_format_dict(data, True)
        cache = open(filename, 'w')
        cache.write(json_data)
        cache.close()

    def to_safe(self, word):
        """Escapes any characters that would be invalid in an ansible group name."""
        return re.sub("[^A-Za-z0-9\-]", "_", word)

    def json_format_dict(self, data, pretty=False):
        """Converts a dict to a JSON object and dumps it as a formatted string."""
        if pretty:
            return json.dumps(data, sort_keys=True, indent=2)
        else:
            return json.dumps(data)


LinodeInventory()
