#!/usr/bin/env python3

'''
Local inventory script
======================

Generates inventory that Ansible can understand by scanning the available
network interfaces.

You need /etc/local.yml with the list of known hosts and their groups
'''

try:
    import configparser
except:
    import ConfigParser as configparser

import argparse
import os
import re
import socket
import yaml
from netifaces import interfaces, ifaddresses, AF_INET
from time import time

import nmap

try:
    import json
except ImportError:
    import simplejson as json


class LocalInventory(object):

    def _empty_inventory(self):
        return {"_meta": {"hostvars": {}}}

    def __init__(self):
        """Main execution path."""

        self.localhost = socket.gethostname()

        # Inventory grouped by display group
        self.inventory = self._empty_inventory()

        # Index of label to Linode ID
        self.index = {}

        # Read settings and parse CLI arguments
        self.parse_cli_args()
        self.read_settings()
        self.read_config()

        # Cache
        if self.args.refresh_cache:
            self.do_scan_update_cache()
        elif not self.is_cache_valid():
            self.do_scan_update_cache()

        # Data to print
        if self.args.host:
            data_to_print = self.get_host_info()
        elif self.args.list:
            # Display list of nodes for inventory
            if self.inventory == self._empty_inventory():
                data_to_print = self.get_inventory_from_cache()
            else:
                data_to_print = self.json_format_dict(self.inventory, True)
        else:
            data_to_print = {}

        print(data_to_print)

    def parse_cli_args(self):
        """Command line argument processing"""
        parser = argparse.ArgumentParser(description='Produce an Ansible Inventory file based on Local network interfaces')
        parser.add_argument('--list', action='store_true', default=True,
                            help='List nodes (default: True)')
        parser.add_argument('--host', action='store',
                            help='Get all the variables about a specific node')
        parser.add_argument('--refresh-cache', action='store_true', default=os.environ.get('ANSIBLE_INVENTORY_REFRESH', False),
                            help='Force refresh of cache by making a new scan (default: False - use cache files)')
        self.args = parser.parse_args()

    def read_settings(self):
        ''' Reads the settings from the local.ini file '''
        config = configparser.ConfigParser()
        local_default_ini_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'local.ini')
        local_ini_path = os.path.expanduser(os.path.expandvars(os.environ.get('LOCAL_INI_PATH', local_default_ini_path)))
        config.read(local_ini_path)

        # Cache related
        cache_dir = os.path.expanduser(config.get('local', 'cache_path'))
        if not os.path.exists(cache_dir):
            os.makedirs(cache_dir)
        self.cache_path_cache = cache_dir + "/ansible-local" + ".cache"
        self.cache_path_index = cache_dir + "/ansible-local" + ".index"
        self.cache_max_age = config.getint('local', 'cache_max_age')
        self.local_ip_mask = config.get('local', 'ip_mask').split(':')

    def read_config(self):
        if not os.path.exists('/etc/local.yml'):
            self.config = {
                'localhost': ['drupal_dev']
            }
        else:
            with open('/etc/local.yml', 'r') as config:
                try:
                    self.config = yaml.load(config, Loader=yaml.BaseLoader)
                except yaml.YAMLError as exc:
                    raise Exception('/etc/local.yaml is no proper YAML file')

    def is_cache_valid(self):
        ''' Determines if the cache files have expired, or if it is still valid '''

        if os.path.isfile(self.cache_path_cache):
            mod_time = os.path.getmtime(self.cache_path_cache)
            current_time = time()
            if (mod_time + self.cache_max_age) > current_time:
                if os.path.isfile(self.cache_path_index):
                    return True

        return False

    def get_host_info(self):
        """Get variables about a specific host."""

        if len(self.index) == 0:
            # Need to load index from cache
            self.load_index_from_cache()

        if not self.args.host in self.index:
            # try updating the cache
            self.do_scan_update_cache()

        if not self.args.host in self.index:
            # host might not exist anymore
            return self.json_format_dict({}, True)

        return self.json_format_dict(self.args.host, True)

    def getLocalname(self, hostname):
        hostname = hostname.split('.')[0]
        if hostname == 'localhost':
            return self.localhost
        return hostname

    def getHostname(self, host):
        for h in host['hostnames']:
            if h['type'] == 'user':
                return self.getLocalname(h['name'])

        if len(host['hostnames']) > 0 and 'name' in host['hostnames'][0]:
            return self.getLocalname(host['hostnames'][0]['name'])

        return self.getLocalname('unknown')

    def getHostGroups(self, hostname):
        if self.config == True:
            return ['everything']
        if hostname not in self.config:
            return False
        return self.config[hostname]

    def do_scan_update_cache(self):
        """Do scan, and save data in cache files."""
        self.get_nodes()
        self.write_to_cache(self.inventory, self.cache_path_cache)
        self.write_to_cache(self.index, self.cache_path_index)

    def _get_local_mask(self, ip):
        parts = ip.split('.')
        parts.pop()
        return '.'.join(parts)

    def get_nodes(self):
        if True or os.environ.get('ANSIBLE_LOCAL_LIMITED', 'False') == 'True':
            hosts = []
            local = self.local_ip_mask
            for host in self.config:
                try:
                    (name, aliaslist, addresslist) = socket.gethostbyname_ex(host)
                    node = {
                        'label': host,
                        'addresslist': addresslist,
                        'interface': 'eth0',
                        'state': 'up'
                    }
                    if '127.0.0.1' in addresslist:
                        for address in addresslist:
                            if address != '127.0.0.1':
                                local.append(self._get_local_mask(address))
                        node['ip'] = '127.0.0.1'
                        self.add_node(node)
                    else:
                        hosts.append(node)
                except socket.gaierror:
                    pass
            for node in hosts:
                for address in node['addresslist']:
                    if self._get_local_mask(address) in local:
                        node['ip'] = address
                        self.add_node(node)
            return


        """Scan local interfaces to get the list of nodes."""
        nm = nmap.PortScanner()
        for ifaceName in interfaces():
            if ifaceName != 'lo':
                for address in [i['addr'] for i in ifaddresses(ifaceName).setdefault(AF_INET, [{'addr':'No IP addr'}] )]:
                    nm.scan(address + '/24', '22')
                    for host in nm.all_hosts():
                        node = {
                            'label': self.getHostname(nm[host]),
                            'ip': host,
                            'interface': ifaceName,
                            'state': nm[host].state()
                        }
                        self.add_node(node)

    def add_node(self, node):
        """Adds an node to the inventory and index."""

        dest = node.get('label')

        if dest == 'unknown':
            return

        groups = self.getHostGroups(dest)
        if node.get('state') == 'up' and groups:
            for group in groups:
                self.push(self.inventory, self.to_safe(group), dest)
        else:
            self.push(self.inventory, 'norootserver', dest)
            self.push(self.inventory, 'offline', dest)

        # Global Tag: tag all JiffyBox instances
        self.push(self.inventory, 'local', dest)

        self.inventory["_meta"]["hostvars"][dest] = self.get_host_info_dict_from_instance(node)

    def get_host_info_dict_from_instance(self, node):
        instance_vars = {}
        for direct_attr in [
            "label",
            "ip",
            "interface",
            "state"
        ]:
            instance_vars["local_" + direct_attr] = node.get(direct_attr)

        instance_vars['static_ipv4'] = instance_vars['local_ip']
        return instance_vars

    def push(self, my_dict, key, element):
        """Pushed an element onto an array that may not have been defined in the dict."""
        if key in my_dict:
            my_dict[key].append(element)
        else:
            my_dict[key] = [element]

    def get_inventory_from_cache(self):
        """Reads the inventory from the cache file and returns it as a JSON object."""
        cache = open(self.cache_path_cache, 'r')
        json_inventory = cache.read()
        return json_inventory

    def load_index_from_cache(self):
        """Reads the index from the cache file and sets self.index."""
        cache = open(self.cache_path_index, 'r')
        json_index = cache.read()
        self.index = json.loads(json_index)

    def write_to_cache(self, data, filename):
        """Writes data in JSON format to a file."""
        json_data = self.json_format_dict(data, True)
        cache = open(filename, 'w')
        cache.write(json_data)
        cache.close()

    def to_safe(self, word):
        """Escapes any characters that would be invalid in an ansible group name."""
        return re.sub("[^A-Za-z0-9\-]", "_", word)

    def json_format_dict(self, data, pretty=False):
        """Converts a dict to a JSON object and dumps it as a formatted string."""
        if pretty:
            return json.dumps(data, sort_keys=True, indent=2)
        else:
            return json.dumps(data)


LocalInventory()
