import json
import os
import shutil
import threading
import yaml
from subprocess import Popen, PIPE, STDOUT

import sys

from lib.mylogqueue import MyLogQueue
from lib.script import Script


class MyAnsible:

    queue = None
    error_seen = False
    guiAvailable = False
    args = None
    extras = None
    guiGiven = False

    try:
        # noinspection PyUnresolvedReferences
        from lib.window import GuiPart
        guiAvailable = True
    except ImportError:
        pass


    def __init__(self, parser, script=Script(False)):
        self.env = os.environ.copy()
        self.parser = parser
        self.script = script


    @staticmethod
    def error(msg):
        sys.stderr.write(msg)
        exit(1)


    def run(self):
        self.parser.add_argument('--limit', default='all',
                            help='Host name or pattern')
        self.parser.add_argument('--debug', action='store_true',
                            help='Turn Debug mode on')
        self.parser.add_argument('--log-path', default=None,
                            help='Path to an existing directory where output should be written to')
        if self.guiAvailable:
            self.parser.add_argument('--gui', action='store_true',
                                help='Enable GUI')
        if not os.path.exists(self.script.rootPath + 'inventory/group_vars') and not self.script.ignoresInventory():
            self.parser.add_argument('--company', default=os.environ.get('ANSIBLE_COMPANY', None),
                                required=self._companyRequired(),
                                help='Only useful if inventories from multiple companies are installed')
            self.parser.add_argument('--include-local', action='store_true',
                                help='If company=all, this would also include the local inventory, default to False')
        else:
            self.parser.add_argument('--company', default=os.environ.get('ANSIBLE_COMPANY', None),
                                help='Allow this argument but it is not really used')

        self.script.argParser(self.parser)

        self.args, self.extras = self.parser.parse_known_args()
        self.args.path = self.script.rootPath

        self.script.setArguments(self.args)

        if self.guiAvailable:
            # Remember this setting for possible followers
            self.guiGiven = self.args.gui
        if self.guiAvailable and (not self.args.gui or self.args.debug):
            self.guiAvailable = False

        if 'playbook' in self.args and 'limit' in self.args:
            self.extras.append('--limit=' + self.args.limit)

        mode = self._getMode()

        # Prepare hostvars cache directory
        cachedir = os.path.expanduser('~/.ansible/hostvars')
        if not os.path.exists(cachedir):
            os.makedirs(cachedir)

        # Check if this is executed from GitLab-Runner and set arguments accordingly
        if mode == 'gitlab-runner':
            company = self._prepareGitLabRunnerMode()
        elif 'company' in self.args:
            company = self.args.company
        else:
            company = None

        if self.script.ignoresInventory():
            # We are running a script that doesn't require the inventory
            thread = MyAnsibleThread(self)
            thread.start()
            thread.join()
        else:
            files = os.listdir(self.script.rootPath + 'inventory')
            if 'group_vars' in files:
                # This is a single inventory installation, call it just once
                thread = MyAnsibleThread(self)
                thread.start()
                thread.join()
            elif not company:
                self.error('No company given')
            else:
                if self.args.log_path:
                    self.guiAvailable = False
                    self.queue = MyLogQueue(self.args.log_path)
                    self._startThreads(company, files)
                elif self.guiAvailable:
                    self.guiAvailable = self.GuiPart()
                    self.queue = self.guiAvailable.queue
                    self._startThreads(company, files)
                    self.guiAvailable.main()
                else:
                    self._startThreads(company, files)

        if self.error_seen:
            self.error('Finished with error!')


    def _prepareGitLabRunnerMode(self):
        if 'CI_PROJECT_ID' not in self.env:
            self.error('Missing project ID from environment')
        deploy_config = os.path.expanduser('~/.config/gitlab-runner/' + self.env['CI_PROJECT_ID'] + '.yml')
        if not os.path.isfile(deploy_config):
            self.error('Missing project configuration')
        deploy = []
        with open(deploy_config, 'r') as deploy_stream:
            try:
                deploy = yaml.load(deploy_stream, Loader=yaml.BaseLoader)
            except yaml.YAMLError:
                self.error('Project configuration is no proper YAML file')
        if 'company' not in deploy:
            self.error('Missing company definition in project configuration')
        if 'host' not in deploy:
            deploy['host'] = 'all'
        self.extras.append('--limit=' + deploy['host'])
        if 'playbook' in deploy:
            self.args.playbook = deploy['playbook']
        if 'extras' in deploy:
            self.extras.append('--extra-vars=' + json.dumps(deploy['extras']))
        if 'tags' in deploy:
            self.extras.append('--tags=' + deploy['tags'])
        else:
            self.extras.append('--tags=deploy')
        if 'start' in deploy:
            self.extras.append('--start-at-task=' + deploy['start'])
        return deploy['company']


    def _startThreads(self, company, files):
        threads = []
        main_started = 0
        companies = company.split(',')
        excluded = ['.git', 'cloud']
        if not self.args.include_local and not 'local' in companies:
            excluded.append('local')
        for filename in files:
            if filename in excluded or os.path.isfile(self.script.rootPath + 'inventory' + os.path.sep + filename):
                continue
            if 'all' in companies or filename in companies:
                main_started += 1
                thread = MyAnsibleThread(self, filename)
                thread.start()
                threads.append(thread)
        if main_started == 0:
            self.error('No valid company given')

        if self.guiAvailable:
            pass
        else:
            if self.queue:
                thread = threading.Thread(target=self.queue.periodicCall)
                thread.start()
            for thread in threads:
                thread.join()

        if self.queue:
            self.queue.finishOutput()


    def _companyRequired(self):
        if 'ANSIBLE_COMPANY' in self.env:
            return False
        if self.script.givesInventoryByScript():
            return False
        return not self._isGitlabUser()


    def _isGitlabUser(self):
        return 'USER' in self.env and self.env['USER'] == 'gitlab-runner'


    def _getMode(self):
        if self._isGitlabUser():
            if 'playbook' in self.args:
                if self.args.playbook == 'setup-local':
                    return 'default'

                elif self.args.playbook != 'deploy':
                    raise Exception('This is not the way to use the Ansible framework')

            return 'gitlab-runner'

        return 'default'


class MyAnsibleThread(threading.Thread):

    def __init__(self, main, company = None):
        # type: (MyAnsible, str) -> None
        """Something to explain what is going on here"""
        threading.Thread.__init__(self)
        self.main = main
        self.company = company


    def run(self):
        # Start building command
        if 'playbook' in self.main.args:
            cmd = []
        else:
            cmd = ['python3', shutil.which('ansible'), self.main.args.limit]

        # Determine inventory file
        if 'local' in self.main.args and self.main.args.local:
            inventory = self.main.script.rootPath + 'local.inventory'
            pathCustom = False
        elif os.path.exists(self.main.script.rootPath + 'inventory/inventory'):
            inventory = self.main.script.rootPath + 'inventory/inventory'
            pathCustom = self.main.script.rootPath + 'inventory/'
        elif self.company and os.path.exists(self.main.script.rootPath + 'inventory/' + self.company + '/inventory'):
            inventory = self.main.script.rootPath + 'inventory/' + self.company + '/inventory'
            pathCustom = self.main.script.rootPath + 'inventory/' + self.company + '/'
        elif self.company and os.path.exists(self.main.script.rootPath + 'inventory/' + self.company):
            inventory = self.main.script.rootPath + 'inventory/' + self.company
            pathCustom = self.main.script.rootPath + 'inventory/' + self.company + '/'
        else:
            inventory = self.main.script.rootPath + 'inventory'
            pathCustom = self.main.script.rootPath + 'inventory/'
        if not os.path.exists(inventory) or (os.path.isdir(inventory) and not os.path.exists(inventory + '/group_vars')):
            self.main.error('Inventory %s not found' % inventory)
        cmd.append('-i')
        cmd.append(inventory)

        # Read user vars if available
        self._prepare_user_vars(cmd, pathCustom)

        # Check for the vaults
        vaultFound = False
        vaults = [
            os.path.expanduser('~/.ansible/secrets'),
            os.path.expanduser('~/.ansible/user/secrets'),
            os.path.expanduser('%s/secrets' % pathCustom)
        ]
        for pathSecrets in vaults:
            if os.path.exists(pathSecrets):
                cmd.append('-e')
                cmd.append('@' + pathSecrets)
                vaultFound = True
        if not vaultFound:
            cmd.append('--ask-become-pass')

        # Check for excluded roles
        if 'roles_exclude' in self.main.args and self.main.args.roles_exclude:
            excluded_roles = self.main.args.roles_exclude.split(',')
        else:
            excluded_roles = []
        cmd.append('--extra-vars={excluded_roles: ' + json.dumps(excluded_roles) + '}')

        self.main.script.setCommand(cmd, self.main.args, self.main.extras)

        # Append more CLI options
        for extra in self.main.extras:
            cmd.append(extra)

        env = os.environ.copy()
        if self.company:
            env['ANSIBLE_COMPANY'] = self.company
            env['FLUENTD_COMPANY'] = self.company
        else:
            # We keep FLUENTD_COMPANY if available
            env.pop('ANSIBLE_COMPANY', None)

        self.main.script.setEnvironment(env)

        if 'playbook' in self.main.args:
            # Add the current scope as an extra variable for tasks to make use of
            cmd.append('--extra-vars={playbook_scope: ' + self.main.args.limit + '}')
            # Run the main playbook if it exists and if not in custom mode
            if (not self.main.args.custom or self.main.args.custom == 'no') and os.path.exists(self.main.script.rootPath + 'playbooks' + os.path.sep + self.main.args.playbook + '.yml'):
                self._play('playbooks' + os.path.sep + self.main.args.playbook, cmd, env)

            # Optionally also run the custom playbook
            if not self.main.error_seen and pathCustom and os.path.exists(pathCustom + self.main.args.playbook + '.yml'):
                self._play(pathCustom + self.main.args.playbook, cmd, env)

        else:
            # Append parameter such that we always run as sudo
            cmd.append('-b')

            # Run Ansible
            self._subprocess(cmd, env)

        for follower in self.main.script.getFollower():
            if self.main.error_seen:
                return

            cmd, condition = self.main.script.prepareFollower(follower, self.main.args, self.main.extras)

            if hasattr(self.main.args, 'company') and self.main.args.company:
                cmd.append('--company=' + self.main.args.company)
            if self.main.args.debug:
                cmd.append('--debug')
            if self.main.args.log_path:
                cmd.append('--log-path=' + self.main.args.log_path)
            if self.main.guiGiven:
                cmd.append('--gui')

            if condition:
                process = self._openProcess(cmd, cwd=self.main.script.rootPath, env=env)
                process.wait()
                if process.returncode > 0:
                    self.main.error_seen = True


    def _prepare_user_vars(self, cmd, path_custom):
        if path_custom and os.path.exists(path_custom + '/user_vars'):
            uservars = {}
            for filename in os.listdir(path_custom + '/user_vars'):
                if filename.endswith('.yml'):
                    uservars.__setitem__(os.path.splitext(os.path.basename(filename))[0], yaml.load(open(path_custom + '/user_vars/' + filename), Loader=yaml.BaseLoader))
            filename = os.path.expanduser('~/.ansible/uservars')
            if not os.path.exists(filename):
                os.makedirs(filename)
            if self.company:
                inventory_name = self.company
            else:
                inventory_name = 'all'
            filename = filename + '/' + inventory_name + '.yml'
            yaml.dump({'users': uservars}, open(filename, 'w'), default_flow_style=False)
            cmd.append('-e')
            cmd.append('@' + filename)
        pathVars = os.path.expanduser('~/.config/ansible.yml')
        if os.path.exists(pathVars):
            cmd.append('-e')
            cmd.append('@' + pathVars)
        pathVars = os.path.expanduser('/etc/ansible.yml')
        if os.path.exists(pathVars):
            cmd.append('-e')
            cmd.append('@' + pathVars)


    def _play(self, playbook, args, env):
        cmd = ['python3', shutil.which('ansible-playbook'), playbook + '.yml']
        for arg in args:
            cmd.append(arg)
        self._subprocess(cmd, env)


    def _subprocess(self, cmd, env):
        if self.main.queue:
            self.main.queue.put({
                'company': env['ANSIBLE_COMPANY'],
                'started': True,
            })

            process = self._openProcess(cmd, cwd=self.main.script.rootPath, env=env, stdout=PIPE, stderr=STDOUT)

            while True:
                out = process.stdout.readline()
                if process.poll() is not None:
                    self.main.queue.put({
                        'company': env['ANSIBLE_COMPANY'],
                        'completed': True,
                        'returncode': process.returncode,
                    })
                    break
                if out != '':
                    self.main.queue.put({
                        'company': env['ANSIBLE_COMPANY'],
                        'msg': out,
                    })
        else:
            process = self._openProcess(cmd, cwd=self.main.script.rootPath, env=env)
            process.wait()

        if process.returncode > 0:
            self.main.error_seen = True


    def _openProcess(self, cmd, **args):
        if self.main.args.debug and cmd[0] != self.main.script.executable():
            cmd.insert(0, 'echo')
            for arg in args:
                cmd.append('[%s] = %s' % (arg, args[arg]))
        return Popen(cmd, **args)
