import os
import sys
import yaml
from six import string_types

import lib.general


class Script:

    ignoreInventory = None
    inventoryGivenByScript = None
    loaded = False


    def __init__(self, load = True):
        self.rootPath = os.path.realpath(os.path.dirname(os.path.realpath(__file__)) + os.path.sep + '..') + os.path.sep
        self.scriptPath = self.rootPath + 'scripts' + os.path.sep
        if load:
            self._loadFromCommandline()


    @staticmethod
    def _extractExtraValue(extras, key):
        label = '--%s=' % key
        for extra in extras:
            if extra.find(label) == 0:
                return extra[len(label):]
        return False


    @staticmethod
    def _buildScriptParser(script_def, parser):
        for parserType in ['arguments', 'options']:
            prefix = '--' if parserType == 'options' else ''
            if parserType in script_def:
                for item in script_def[parserType]:
                    name = item['name']
                    item.pop('name')
                    for kw in item:
                        if isinstance(item[kw], dict) and 'exec' in item[kw]:
                            item[kw] = eval(item[kw]['exec'])
                    parser.add_argument(prefix + name, **item)


    def ignoresInventory(self):
        if self.ignoreInventory is None:
            self.ignoreInventory = (self.loaded and 'ignore-inventory' in self.scr and self.scr['ignore-inventory'])
        return self.ignoreInventory


    def givesInventoryByScript(self):
        if self.inventoryGivenByScript is None:
            self.inventoryGivenByScript = (self.loaded and 'inventory-given-by-script' in self.scr and self.scr['inventory-given-by-script'])
        return self.inventoryGivenByScript


    def executable(self):
        return self.rootPath + 'ansible-script.py'


    def argParser(self, parser):
        if not self.loaded:
            return
        parser.description = self.scr['description']
        if 'cli' in self.scr:
            for groupflag in [None, 'groups']:
                if groupflag is None:
                    self._buildScriptParser(self.scr['cli'], parser)
                elif groupflag in self.scr['cli']:
                    for group in self.scr['cli'][groupflag]:
                        parserGroup = parser.add_argument_group(group, self.scr['cli'][groupflag][group]['description'])
                        self._buildScriptParser(self.scr['cli'][groupflag][group], parserGroup)
            if 'defaults' in self.scr['cli']:
                for item in self.scr['cli']['defaults']:
                    parser.set_defaults(**{item: self.scr['cli']['defaults'][item]})


    def setArguments(self, args):
        if not self.loaded:
            return
        if 'arguments' not in self.scr:
            return
        for item in self.scr['arguments']:
            value = self.scr['arguments'][item]
            if not isinstance(value, string_types):
                value = ','.join(value)
            setattr(args, item, value)
        if hasattr(args, 'playbook'):
            if not hasattr(args, 'custom'):
                setattr(args, 'custom', False)


    def setCommand(self, cmd, args, extras):
        if not self.loaded:
            return
        if 'command' not in self.scr:
            return
        self._buildCommand(cmd, self.scr['command'], args, extras)


    def setEnvironment(self, env):
        if not self.loaded:
            return
        if 'environment' not in self.scr:
            return
        for item in self.scr['environment']:
            env[item] = self.scr['environment'][item]


    def getFollower(self):
        if not self.loaded:
            return []
        if 'follower' not in self.scr:
            return []
        return self.scr['follower']


    def prepareFollower(self, follower, args, extras):
        if not isinstance(follower, string_types):
            condition = self._evaluateCondition(follower, args, extras)
            cmd = [self.executable(), follower['script']]
            if 'arguments' in follower:
                for argument in follower['arguments']:
                    cmd.append(argument)
            if 'command' in follower:
                self._buildCommand(cmd, follower['command'], args, extras)
        else:
            condition = True
            cmd = [self.executable(), follower, '--limit=' + args.limit]

        if 'negate' in follower and follower['negate']:
            condition = not condition

        return cmd, condition


    def _loadFromCommandline(self):
        l = len(sys.argv)
        if l == 1:
            self._listAllScripts()
            exit(0)

        arg = sys.argv[1]
        if arg[0] == '-':
            lib.general.MyAnsible.error('First arguments needs to be a script name.')

        if arg == 'list':
            self._listAllScripts()
            exit(0)

        filename = self.scriptPath + arg + '.yml'

        if not os.path.exists(filename):
            lib.general.MyAnsible.error('Script %s does not exist.' % arg)

        with open(filename, 'r') as config:
            try:
                self.scr = yaml.load(config, Loader=yaml.BaseLoader)
                self.loaded = True
            except yaml.YAMLError:
                lib.general.MyAnsible.error('Script %s can not be loaded.' % arg)


    def _listAllScripts(self):
        scripts = []
        for filename in sorted(os.listdir(self.scriptPath)):
            name, ext = os.path.splitext(filename)
            if ext == '.yml':
                with open(self.scriptPath + filename, 'r') as config:
                    try:
                        script = yaml.load(config, Loader=yaml.BaseLoader)
                        scripts.append([name, script['description']])
                    except yaml.YAMLError:
                        pass
        try:
            # noinspection PyUnresolvedReferences
            from tabulate import tabulate
            print(tabulate(scripts, headers=['Script', 'Description']))
        except ImportError:
            print(scripts)


    def _buildCommand(self, cmd, command_def, args, extras):
        for item in command_def:
            condition = self._evaluateCondition(item, args, extras)
            if 'negate' in item and item['negate']:
                condition = not condition
            if condition:
                if 'value' in item:
                    value = item['value']
                    if not isinstance(value, string_types):
                        value = ','.join(value)
                    if 'args' in item:
                        commandargs = []
                        for commandarg in item['args']:
                            if isinstance(commandarg, dict) and 'exec' in commandarg:
                                commandarg = commandarg['exec']['name']
                                setattr(args, commandarg, eval(commandarg['exec']['value']))
                            if hasattr(args, commandarg):
                                commandargs.append(getattr(args, commandarg))
                            else:
                                commandargs.append(self._extractExtraValue(extras, commandarg))
                        value = value % tuple(commandargs)
                    cmd.append('--' + item['name'] + '=' + value)
                else:
                    cmd.append('--' + item['name'])


    def _evaluateCondition(self, item, args, extras):
        condition = True
        if 'condition' in item:
            if isinstance(item['condition'], dict) and 'exec' in item['condition']:
                condition = eval(item['condition']['exec'])
            elif hasattr(args, item['condition']):
                condition = bool(getattr(args, item['condition']))
            else:
                condition = bool(self._extractExtraValue(extras, item['condition']))

        return condition
